/*
	FVCool version 1.02

	VCool for FreeBSD (and possibly Liunx):
	CPU cooling software for AMD's Athlon/Duron motherboard.

 */

/* Original is VCool of Marin Peters in the following */
/***************************************************************************
                          LVCool.cpp
 Sets the "enable Bus disconnect on STPGNT" bit on the northbridge
 and runs and idle loop that puts the CPU into STPGNT mode.

 Since I'm not a Linux guy just the minimum approach here *g*


					-------------------
	begin			: Fri Jul  6 10:13:24 CEST 2001
	copyright		: (C) 2001 by Martin Peters
	email			: mpet@bigfoot.de
	URL				: http://vcool.occludo.net/
 ***************************************************************************/

/* Data for Athlon/Duron Cooling information by Hidemi Oya */
/***************************************************************************
                          CoolON
	URL				: http://homepage2.nifty.com/coolon/ (Japanese)
 ***************************************************************************/

/*
	<<< BASIC DATASHEETS >>>
	disconnect CPU bus at HLT or STPGNT state: bit in Northbridge

	AMD 751/761/762
		 PCI conf. registor 0x62  bit 1 and bit2    i.e. != 0x06
		(PCI conf. registor 0x60  bit 17 and bit18  i.e. != 0x060000)

	Sis 730
		 PCI conf. registor 0x6B  bit 0   i.e. != 0x01
		(PCI conf. registor 0x68  bit 24  i.e. != 0x01000000)

	Sis 735
		 PCI conf. registor 0x6A  bit 0   i.e. != 0x01
		(PCI conf. registor 0x68  bit 16  i.e. != 0x010000)

	VIA KT400
		 PCI conf. registor 0xD2  bit 7   i.e. != 0x80      < --- STPGNT
		(PCI conf. registor 0xD0  bit 23  i.e. != 0x800000)
		 PCI conf. registor 0xD5  bit 1   i.e. != 0x02      < --- HLT
		(PCI conf. registor 0xD4  bit 9   i.e. != 0x0200)

	VIA KT266/A/333
		 PCI conf. registor 0x92  bit 7   i.e. != 0x80      < --- STPGNT
		(PCI conf. registor 0x90  bit 23  i.e. != 0x800000)
		 PCI conf. registor 0x95  bit 1   i.e. != 0x02      < --- HLT
		(PCI conf. registor 0x94  bit 9   i.e. != 0x0200)

	VIA KT133/A/KX133
		 PCI conf. registor 0x52  bit 7   i.e. != 0x80      < --- STPGNT
		(PCI conf. registor 0x50  bit 23  i.e. != 0x800000)
	with additional ACPI operation: set STPGNT state in idle-loop.
	This is done by reading ACPI Processor Level 2 registor.

 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#ifndef LINUX	/* FreeBSD */
#include <machine/cpufunc.h>
#define OUTl(x,y)	outl((x),(y))
#define WAIT	outb(0xEB,0x00)
#else			/* LINUX */
#include <sys/io.h>
#define OUTl(x,y)	outl((y),(x))
#define WAIT	outb(0x00,0xEB)
#endif

#ifndef LINUX	/* FreeBSD */
static int iofl;
#else			/* LINUX */
int iopl_counter = 0;
#endif

static int OpenIO()
{
#ifndef LINUX	/* FreeBSD */
	if ((iofl = open("/dev/io",000)) < 0)
		return -1;
#else			/* LINUX */
	if (iopl_counter == 0) {
		if (iopl(3) < 0)
			return -1;
	}
	++iopl_counter;
#endif
	return 0;
}
static void CloseIO()
{
#ifndef LINUX	/* FreeBSD */
	close(iofl);
#else			/* LINUX */
	if (iopl_counter == 1)
		iopl(0);
	if (iopl_counter > 0)
		--iopl_counter;
#endif
}

#define DWORD unsigned int

#define PCI_CFIO	0x0CF8
#define PCI_CWRV	0x0CFC
#define PCI_BUSM	0x03
#define PCI_DEVM	0x32
#define PCI_FUNM	0x08

/* supported chipsets */

#define ID_AMD751	0x70061022
#define ID_AMD751S	0x70041022
#define ID_AMD761	0x700E1022
#define ID_AMD762	0x700C1022
#define AMD_7XX		0x01

#define ID_VIA8363	0x03051106
#define ID_VIA8371	0x03911106
#define ID_VIA8366	0x30991106
#define ID_VIA8368	0x31891106
#define ID_VIA686	0x30571106
#define VIA_KT133	0x11
#define VIA_KT266	0x12
#define VIA_KT400	0x13

#define ID_SIS730	0x07301039
#define ID_SIS735	0x07351039
#define SIS_730		0x21
#define SIS_735		0x22

/* global variables */
int Reg_PL2;
int North_Chip = 0;
int idle_flag = 0, debug_flag = 0, enable_flag = 0;

DWORD PCIRead(int reg, int fun, int dev, int bus)
{
	DWORD r = 0x80000000, ret, org;
	
	r |= (( bus & 0xff) <<16);
	r |= (( dev & 0x1f) <<11);
	r |= (( fun & 0x07) << 8);
	r |= (( reg & 0xfc) );
	
	org = inl(PCI_CFIO); WAIT;
	OUTl(PCI_CFIO, r); WAIT;
	ret = inl(PCI_CWRV); WAIT;
	org &= 0x7FFFFFFF;
	OUTl(PCI_CFIO, org); WAIT;
	
	return ret;
}

void PCIWrite(DWORD val, int reg, int fun, int dev, int bus)
{
	DWORD r = 0x80000000, org;
	
	r |= (( bus & 0xff) <<16);
	r |= (( dev & 0x1f) <<11);
	r |= (( fun & 0x07) << 8);
	r |= (( reg & 0xfc) );

	org = inl(PCI_CFIO); WAIT;
	OUTl(PCI_CFIO, r); WAIT;
	OUTl(PCI_CWRV, val); WAIT;
	org &= 0x7FFFFFFF;
	OUTl(PCI_CFIO, org); WAIT;
}

/* scan PCI bus for north and south (VIA82C686 only) bridge */

int search_PCI(int *chip, int *nb, int *nd, int *nf, int *sb, int *sd, int *sf)
{
	int bus, dev, fun, res, ret = 0;

	*chip = 0;
	for (bus = 0; bus < PCI_BUSM; bus++) {
		for (dev = 0; dev < PCI_DEVM; dev++) {
			for (fun = 0; fun < PCI_FUNM; fun++) {
				res = PCIRead(0, fun, dev, bus);
				if (res == 0xffffffff)
					continue;
				switch (res) {
				case ID_AMD751:
				case ID_AMD751S:
				case ID_AMD761:
				case ID_AMD762:
					*chip = AMD_7XX;
					break;
				case ID_VIA8363:
				case ID_VIA8371:
					*chip = VIA_KT133;
					break;
				case ID_VIA8366:
					*chip = VIA_KT266;
					break;
				case ID_VIA8368:
				  *chip = VIA_KT400;
				  break;

				case ID_SIS730:
					*chip = SIS_730;
					break;
				case ID_SIS735:
					*chip = SIS_735;
					break;
				case ID_VIA686:
					*sb = bus;
					*sd = dev;
					*sf = fun;
					ret |= 0x02;
					break;
				}
				if (!*chip) {
					*nb = bus;
					*nd = dev;
					*nf = fun;
					ret |= 0x01;
				}
				if (ret >= 3) break;
			}
			if (ret >= 3) break;
		}
		if (ret >= 3) break;
	}
	return ret;
}

int maskbits(int value, int mask)
{
	int ret;

	if (enable_flag > 0)
		ret = value | mask;
	else if (enable_flag < 0)
		ret = value & ~mask;
	else
		ret = value;

	return ret;
}

void chip_info(int value, char *comment, int reg, int mask)
{
	int val1, val2, shiftbits;

	shiftbits = (reg % 4) * 8;
	val1 = (0xFF & (value >> shiftbits));
	val2 = maskbits(val1, mask);
	fprintf(stderr, "%s", comment);
	fprintf(stderr, " Change Reg0x%02X 0x%02X --> 0x%02X\n", reg, val1, val2);
}

int InitPCI(debug_flag)
{
	DWORD res;
	int ret;
	int nb_b = 0, nb_d = 0, nb_f = 0;
	int sb_b = 0, sb_d = 7, sb_f = 4;
	
	ret = search_PCI(&North_Chip, &nb_b, &nb_d, &nb_f, &sb_b, &sb_d, &sb_f);
	if (ret & 0x01) {
		perror("Supported Athlon/Duron chipset (north bridge) not found");
		exit(1);
	}

/* now set "enable bus Diconnect when HLT/STPGNT detected */

	switch (North_Chip) {
	case AMD_7XX:
		res = PCIRead(0x60, nb_f, nb_d, nb_b);
		PCIWrite(maskbits(res, 0x060000), 0x60, nb_f, nb_d, nb_b);
		if (debug_flag)
			chip_info(res, "AMD751/761/762 found.\n", 0x62, 0x06);
		break;
	case VIA_KT133:
		res = PCIRead(0x50, nb_f, nb_d, nb_b);
		PCIWrite(maskbits(res, 0x800000), 0x50, nb_f, nb_d, nb_b);
		if (debug_flag)
			chip_info(res, "VIA KT133/A/KX133 found.\n", 0x52, 0x80);
		break;
	case VIA_KT266:
		res = PCIRead(0x90, nb_f, nb_d, nb_b);
		PCIWrite(maskbits(res, 0x800000), 0x90, nb_f, nb_d, nb_b);
		if (debug_flag)
			chip_info(res, "VIA KT266/A/KT333 found.\n", 0x92, 0x80);
		res = PCIRead(0x94, nb_f, nb_d, nb_b);
		PCIWrite(maskbits(res, 0x0200), 0x94, nb_f, nb_d, nb_b);
		if (debug_flag)
			chip_info(res, "", 0x95, 0x02);
		break;
	case VIA_KT400:
		res = PCIRead(0xd0, nb_f, nb_d, nb_b);
		PCIWrite(maskbits(res, 0x800000), 0xD0, nb_f, nb_d, nb_b);
		if (debug_flag)
			chip_info(res, "VIA KT400 found.\n", 0xD2, 0x80);

		res = PCIRead(0xd4, nb_f, nb_d, nb_b);
		PCIWrite(maskbits(res, 0x0200), 0xD4, nb_f, nb_d, nb_b);
		if (debug_flag)
			chip_info(res, "", 0xD5, 0x02);
		break;
	case SIS_730:
		res = PCIRead(0x68, nb_f, nb_d, nb_b);
		PCIWrite(maskbits(res, 0x01000000), 0x68, nb_f, nb_d, nb_b);
		if (debug_flag)
			chip_info(res, "SiS730 found.\n", 0x6B, 0x01);
		break;
	case SIS_735:
		res = PCIRead(0x68, nb_f, nb_d, nb_b);
		PCIWrite(maskbits(res, 0x010000), 0x68, nb_f, nb_d, nb_b);
		if (debug_flag)
			chip_info(res, "SiS735 found.\n", 0x6A, 0x01);
		break;
	}

/* enable ACPI and find I/O-Space for VT82C686(KT133/A south) */

	if ((ret & 0x02) && (enable_flag > 0)) {
		res = PCIRead(0x41,sb_f,sb_d,sb_b);
		if ((res & 0x8000) == 0) {
			res |= 0x8000;
			PCIWrite(res, 0x41, sb_f, sb_d, sb_b);
		}

		res = PCIRead(0x48, sb_f, sb_d, sb_b);
		res &= 0xff80;

		Reg_PL2 = res + 0x14;

		if (debug_flag) {
			fprintf(stderr, "VT82C686/A/B(ACPI) found.\n");
			fprintf(stderr, " PLVL_2 Reg Address:0x%0X\n", Reg_PL2);
		}
	}
	return ret;
}

void Idleloop()
{
	if(nice(20) == -1) {
		perror("nice does not work");
		exit(1);
	}
	while(1) {
		inb(Reg_PL2);
/*		usleep(30);	*/
	}
}

void usage() {
	fprintf(stderr, "Usage: [-v (verbose)] "
		"[ -e (enable) [ -i (idle loop)]] [ -d (disable)]\n");

	exit(1);
}

int main(int argc, char *argv[])
{
	int n, ch;
	int arguments_processed = 0;

	while ((ch = getopt(argc, argv, "vidhe")) != -1) {
		switch (ch) {
		case 'v':
			debug_flag = 1;
			break;
		case 'i':
			idle_flag = 1;
			break;
		case 'd':
			enable_flag = -1;
			break;
		case 'e':
			enable_flag = 1;
			break;
		case 'h':
		default:
			usage();
		}
		arguments_processed++;
	}

	if (arguments_processed == 0)
		usage();
	
	if (enable_flag <= 0)
		idle_flag = 0;

 	if (OpenIO() < 0) {
		perror("cannot open IO-port, must run as root");
		exit(1);
	}

	if (idle_flag & !enable_flag) {
		fprintf(stderr, "cannot idle without enabling\n");
		usage();
	}

	n = InitPCI(debug_flag);
	if (idle_flag && (n & 0x02)) {
		fprintf(stderr, "Now going into idle-loop!\n");
		Idleloop();
	}

	CloseIO();
	return 0;
}
